"""
Package-wide settings
"""

__author__ = "Janek Sendrowski"
__contact__ = "sendrowski.janek@gmail.com"
__date__ = "2023-09-12"


class Settings:
    """
    Class that holds package-wide settings
    """
    #: Whether to disable the progress bar.
    disable_pbar = False
